/*
 * main.c
 *
 * Created: 05/23/2019 2:19:11 AM
 * Author : Adrian
 */ 

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <time.h>

#include "lcd/lcd.h"
#include "light_ws2812/light_ws2812.h"

#define MAX_PIX 54
#define COLOR_LEN 6

#define  FACE_NUM 6
#define  LINE_LEN 3
#define  COLM_LEN 3

typedef struct cRGB led_t;
typedef unsigned char byte_t;

byte_t PA1_old = 0;
byte_t PA2_old = 0;
byte_t PA3_old = 0;
byte_t PA4_old = 0;
volatile byte_t secs = 0, mins = 0;
byte_t face = 2;
led_t colors[COLOR_LEN];
led_t leds[MAX_PIX];
byte_t cube[FACE_NUM][LINE_LEN][COLM_LEN];
byte_t dup[FACE_NUM][LINE_LEN][COLM_LEN];

void translation();
char *get_face();
void cube_init();
void scramble();

void duplicate();
void look_right();
void look_down();

void r();
void ri();

void l();
void li();

void u();
void ui();

void d();
void di();

void next();

ISR(TIMER1_COMPA_vect) 
{
	secs++;
	if (secs == 60) {
		secs = 0;
		mins++;
	}
}

/*Start/Stop & Switch Face*/
ISR(PCINT0_vect)
{
	uint16_t timer = 0;
	while(bit_is_clear(PINA, PA0)) {
			timer++;
			_delay_ms(1);
	}
	if(timer > 75) {
		if(timer < 700) {
			next();
			translation();
		} else {
			PA1_old = 0;
			PA2_old = 0;
			PA3_old = 0;
			PA4_old = 0;

			LCD_writeInstr(LCD_INSTR_clearDisplay);
			LCD_printAt(0x00, "****NEW GAME****");
			LCD_printAt(0x40, "****************");
			cube_init();
			translation();
			ws2812_sendarray((uint8_t *) leds, MAX_PIX * 3);
			_delay_ms(3000);
			scramble();
			translation();
			secs = 0;
			mins = 0;
			face = 2;
			LCD_writeInstr(LCD_INSTR_clearDisplay);
			TCNT1 = 0x00;
		}
	}
}

void init_timer()
{
	OCR1A = 0xF424;
	
	TIMSK1 = _BV(OCIE1A);
	TCCR1B = _BV(CS12) | _BV(WGM12);
}

void init_interrupt_button()
{
	PCMSK0 |= _BV(PCINT0);
	PCICR  |= _BV(PCIE0);
}

int main(void)
{
	char face_str[15];
	char cron_str[13];

	/*Colors Initialization*/
	colors[0].r = 125; colors[0].g = 000; colors[0].b = 000; /*RED*/
	colors[1].r = 000; colors[1].g = 125; colors[1].b = 000; /*GREEN*/
	colors[2].r = 000; colors[2].g = 000; colors[2].b = 125; /*BLUE*/
	colors[3].r = 125; colors[3].g = 125; colors[3].b = 125; /*WHITE*/
	colors[4].r = 125; colors[4].g = 000; colors[4].b = 125; /*MAGENTA*/
	colors[5].r = 125; colors[5].g = 125; colors[5].b = 000; /*YELLOW*/

	/*LCD Initialization*/
	DDRC |= _BV(PC2);
	PORTC |= _BV(PC2);
	LCD_init();

	/*NEW GAME*/
	LCD_printAt(0x00, "****NEW GAME****");
	LCD_printAt(0x40, "****************");
	
	cube_init();
	translation();
	ws2812_sendarray((uint8_t *) leds, MAX_PIX * 3);
	
	_delay_ms(3000);
	
	scramble();
	translation();
	
	LCD_writeInstr(LCD_INSTR_clearDisplay);

	/*Buttons Initialization*/
	DDRA &= ~_BV(PA0);
	DDRA &= ~_BV(PA1);
	DDRA &= ~_BV(PA2);
	DDRA &= ~_BV(PA3);
	DDRA &= ~_BV(PA4);

	PORTA |= _BV(PA0)
				| _BV(PA1)
				| _BV(PA2)
				| _BV(PA3)
				| _BV(PA4);

	/*Timer Initialization*/
	init_timer();
	init_interrupt_button();
	sei();

  while (1) {

		/*LEFT*/
		if ((PINA & (1 << PA1)) == 0) {
			if (PA2_old == 1 || PA3_old == 1) {
				if (PA2_old == 1)
				{
					di();
					translation();
				}
				if (PA3_old == 1)
				{
					u();
					translation();
				}
				PA1_old = 0;
				PA2_old = 0;
				PA3_old = 0;
				PA4_old = 0;
			} else {
				PA1_old = 1;
				PA4_old = 0;
			}
			_delay_ms(200);
		}

		/*DOWN*/
		if ((PINA & (1 << PA2)) == 0) {
			if (PA1_old == 1 || PA4_old == 1) {
				if (PA1_old == 1)
				{
					l();
					translation();
				}
				if (PA4_old == 1)
				{
					ri();
					translation();
				}
				PA1_old = 0;
				PA2_old = 0;
				PA3_old = 0;
				PA4_old = 0;
			} else {
				PA2_old = 1;
				PA3_old = 0;
			}
			_delay_ms(200);
		}

		/*UP*/
		if ((PINA & (1 << PA3)) == 0) {
			if (PA1_old == 1 || PA4_old == 1) {
				if (PA1_old == 1)
				{
					li();
					translation();
				}
				if (PA4_old == 1)
				{
					r();
					translation();
				}
				PA1_old = 0;
				PA2_old = 0;
				PA3_old = 0;
				PA4_old = 0;
			} else {
				PA2_old = 0;
				PA3_old = 1;
			}
			_delay_ms(200);
		}

		/*RIGHT*/
		if ((PINA & (1 << PA4)) == 0) {
			if (PA2_old == 1 || PA3_old == 1) {
				if (PA2_old == 1)
				{
					d();
					translation();
				}
				if (PA3_old == 1)
				{
					ui();
					translation();
				}
				PA1_old = 0;
				PA2_old = 0;
				PA3_old = 0;
				PA4_old = 0;
			} else {
				PA1_old = 0;
				PA4_old = 1;
			}
			_delay_ms(200);
		}

		/*Print on LCD*/
		sprintf(face_str, "FACE : %s", get_face());
		if (mins < 10 && secs < 10)
			sprintf(cron_str, "TIME : 0%d:0%d", mins, secs);
		else if (mins < 10 && secs > 9)
			sprintf(cron_str, "TIME : 0%d:%d", mins, secs);
		else if (mins > 9 && secs < 10)
			sprintf(cron_str, "TIME : %d:0%d", mins, secs);
		else
			sprintf(cron_str, "TIME : %d:%d", mins, secs);
		LCD_printAt(0x00, face_str);
		LCD_printAt(0x40, cron_str);

		/*Print on WS2812*/
		ws2812_sendarray((uint8_t *) leds, MAX_PIX * 3);
  }
}

/*Initialization of cube and duplicate*/
void cube_init()
{
	byte_t i, j, k;

	for (i = 0; i < FACE_NUM; ++i)
		for(j = 0; j < LINE_LEN; ++j)
			for(k = 0; k < COLM_LEN; ++k) {
				cube[i][j][k] = i;
				dup[i][j][k] = 0;
			}	
}

/*Get string face*/
char *get_face()
{
	switch (face)
	{
	case 0:
		return "RED";
	case 1:
		return "GREEN";
	case 2:
		return "BLUE";
	case 3:
		return "WHITE";
	case 4:
		return "MAGENTA";
	case 5:
		return "YELLOW";
	default:
		return "";
	}
}

/*Map the cube to the led band*/
void translation()
{
	leds[0] = colors[cube[2][2][0]];
	leds[1] = colors[cube[2][2][1]];
	leds[2] = colors[cube[2][2][2]];
	leds[3] = colors[cube[2][1][2]];
	leds[4] = colors[cube[2][1][1]];
	leds[5] = colors[cube[2][1][0]];
	leds[6] = colors[cube[2][0][0]];
	leds[7] = colors[cube[2][0][1]];
	leds[8] = colors[cube[2][0][2]];

	leds[9] = colors[cube[0][2][2]];
	leds[10] = colors[cube[0][2][1]];
	leds[11] = colors[cube[0][2][0]];
	leds[12] = colors[cube[0][1][0]];
	leds[13] = colors[cube[0][1][1]];
	leds[14] = colors[cube[0][1][2]];
	leds[15] = colors[cube[0][0][2]];
	leds[16] = colors[cube[0][0][1]];
	leds[17] = colors[cube[0][0][0]];

	leds[18] = colors[cube[4][0][2]];
	leds[19] = colors[cube[4][0][1]];
	leds[20] = colors[cube[4][0][0]];
	leds[21] = colors[cube[4][1][0]];
	leds[22] = colors[cube[4][1][1]];
	leds[23] = colors[cube[4][1][2]];
	leds[24] = colors[cube[4][2][2]];
	leds[25] = colors[cube[4][2][1]];
	leds[26] = colors[cube[4][2][0]];

	leds[27] = colors[cube[5][2][2]];
	leds[28] = colors[cube[5][2][1]];
	leds[29] = colors[cube[5][2][0]];
	leds[30] = colors[cube[5][1][0]];
	leds[31] = colors[cube[5][1][1]];
	leds[32] = colors[cube[5][1][2]];
	leds[33] = colors[cube[5][0][2]];
	leds[34] = colors[cube[5][0][1]];
	leds[35] = colors[cube[5][0][0]];

	leds[36] = colors[cube[1][2][2]];
	leds[37] = colors[cube[1][2][1]];
	leds[38] = colors[cube[1][2][0]];
	leds[39] = colors[cube[1][1][0]];
	leds[40] = colors[cube[1][1][1]];
	leds[41] = colors[cube[1][1][2]];
	leds[42] = colors[cube[1][0][2]];
	leds[43] = colors[cube[1][0][1]];
	leds[44] = colors[cube[1][0][0]];

	leds[45] = colors[cube[3][0][2]];
	leds[46] = colors[cube[3][0][1]];
	leds[47] = colors[cube[3][0][0]];
	leds[48] = colors[cube[3][1][0]];
	leds[49] = colors[cube[3][1][1]];
	leds[50] = colors[cube[3][1][2]];
	leds[51] = colors[cube[3][2][2]];
	leds[52] = colors[cube[3][2][1]];
	leds[53] = colors[cube[3][2][0]];
}

/*Create duplicate*/
void duplicate()
{
	byte_t i, j, k;
	
	for (i = 0; i < FACE_NUM; ++i)
		for(j = 0; j < LINE_LEN; ++j)
			for(k = 0; k < COLM_LEN; ++k)
				dup[i][j][k] = cube[i][j][k];
}

/*Scramble the cube*/
void scramble()
{
	byte_t i;

	srand(time(NULL));
	for (i = 0; i < 15; ++i)
	{
		uint8_t num = rand() % 8 + 1;
		switch (num)
		{
			case 1:
				r();
				break;
			case 2:
				ri();
				break;
			case 3:
				u();
				break;
			case 4:
				ui();
				break;
			case 5:
				l();
				break;
			case 6:
				li();
				break;
			case 7:
				d();
				break;
			case 8:
				di();
				break;
		}
		translation();
		ws2812_sendarray((uint8_t *) leds, MAX_PIX * 3);
		_delay_ms(500);

	}
}

/*Switch to right face*/
void look_right()
{
	byte_t i, j;

	duplicate();
	for (i = 0; i < COLM_LEN; ++i)
		for (j = 0; j < LINE_LEN; ++j) {
			cube[1][j][i] = dup[2][j][i];
			cube[4][j][i] = dup[1][j][i];
			cube[3][j][i] = dup[4][j][i];
			cube[2][j][i] = dup[3][j][i];
		}

	cube[0][0][0] = dup[0][2][0];
	cube[0][0][2] = dup[0][0][0];
	cube[0][2][2] = dup[0][0][2];
	cube[0][2][0] = dup[0][2][2];

	cube[0][0][1] = dup[0][1][0];
	cube[0][1][2] = dup[0][0][1];
	cube[0][2][1] = dup[0][1][2];
	cube[0][1][0] = dup[0][2][1];

	cube[5][2][0] = dup[5][0][0];
	cube[5][0][0] = dup[5][0][2];
	cube[5][0][2] = dup[5][2][2];
	cube[5][2][2] = dup[5][2][0];

	cube[5][1][0] = dup[5][0][1];
	cube[5][0][1] = dup[5][1][2];
	cube[5][1][2] = dup[5][2][1];
	cube[5][2][1] = dup[5][1][0];
}

/*Switch to bottom face*/
void look_down()
{
	byte_t i, j;
	
	duplicate();
	for (i = 0; i < LINE_LEN; ++i)
		for (j = 0; j <COLM_LEN; ++j) {
			cube[2][i][2 - j] = dup[5][i][2 - j];
			cube[0][i][2 - j] = dup[2][i][2 - j];
			cube[4][i][j] 	  = dup[0][2 - i][2 - j];
			cube[5][i][2 - j] = dup[4][2 - i][j];
		}

	cube[3][0][2] = dup[3][0][0];
	cube[3][2][2] = dup[3][0][2];
	cube[3][2][0] = dup[3][2][2];
	cube[3][0][0] = dup[3][2][0];
	
	cube[3][1][2] = dup[3][0][1];
	cube[3][2][1] = dup[3][1][2];
	cube[3][1][0] = dup[3][2][1];
	cube[3][0][1] = dup[3][1][0];	
	
	cube[1][0][0] = dup[1][0][2];
	cube[1][0][2] = dup[1][2][2];
	cube[1][2][2] = dup[1][2][0];
	cube[1][2][0] = dup[1][0][0];

	cube[1][0][1] = dup[1][1][2];
	cube[1][1][2] = dup[1][2][1];
	cube[1][2][1] = dup[1][1][0];
	cube[1][1][0] = dup[1][0][1];
}

/*Right clockwise*/
void r()
{
	byte_t i;
	
	duplicate();
	for (i = 0; i < LINE_LEN; ++i) {
		cube[2][i][2] = dup[5][i][2];
		cube[0][i][2] = dup[2][i][2];
		cube[4][i][0] = dup[0][2 - i][2];
		cube[5][i][2] = dup[4][2 - i][0];
	}

	cube[3][0][2] = dup[3][0][0];
	cube[3][2][2] = dup[3][0][2];
	cube[3][2][0] = dup[3][2][2];
	cube[3][0][0] = dup[3][2][0];
	
	cube[3][1][2] = dup[3][0][1];
	cube[3][2][1] = dup[3][1][2];
	cube[3][1][0] = dup[3][2][1];
	cube[3][0][1] = dup[3][1][0];
}

/*Right counter clockwise*/
void ri()
{
	byte_t i;

	duplicate();
	for (i = 0; i < LINE_LEN; ++i) {
		cube[5][i][2] = dup[2][i][2];
		cube[2][i][2] = dup[0][i][2];
		cube[0][2 - i][2] = dup[4][i][0];
		cube[4][2 - i][0] = dup[5][i][2];
	}

	cube[3][0][0] = dup[3][0][2];
	cube[3][0][2] = dup[3][2][2];
	cube[3][2][2] = dup[3][2][0];
	cube[3][2][0] = dup[3][0][0];

	cube[3][0][1] = dup[3][1][2];
	cube[3][1][2] = dup[3][2][1];
	cube[3][2][1] = dup[3][1][0];
	cube[3][1][0] = dup[3][0][1];
}

/*Left clockwise*/
void l()
{
	byte_t i;

	duplicate();
	for (i = 0; i < LINE_LEN; ++i) {
		cube[5][i][0] = dup[2][i][0];
		cube[2][i][0] = dup[0][i][0];
		cube[0][i][0] = dup[4][2 - i][2];
		cube[4][2 - i][2] = dup[5][i][0];
	}

	cube[1][0][2] = dup[1][0][0];
	cube[1][2][2] = dup[1][0][2];
	cube[1][2][0] = dup[1][2][2];
	cube[1][0][0] = dup[1][2][0];

	cube[1][1][2] = dup[1][0][1];
	cube[1][2][1] = dup[1][1][2];
	cube[1][1][0] = dup[1][2][1];
	cube[1][0][1] = dup[1][1][0];
}

/*Left counter clockwise*/
void li()
{
	byte_t i;

	duplicate();
	for (i = 0; i < LINE_LEN; ++i) {
		cube[2][i][0] = dup[5][i][0];
		cube[0][i][0] = dup[2][i][0];
		cube[4][i][2] = dup[0][2 - i][0];
		cube[5][i][0] = dup[4][2 - i][2];
	}

	cube[1][0][0] = dup[1][0][2];
	cube[1][0][2] = dup[1][2][2];
	cube[1][2][2] = dup[1][2][0];
	cube[1][2][0] = dup[1][0][0];

	cube[1][0][1] = dup[1][1][2];
	cube[1][1][2] = dup[1][2][1];
	cube[1][2][1] = dup[1][1][0];
	cube[1][1][0] = dup[1][0][1];
}

/*Up clockwise*/
void u()
{
	byte_t i;

	duplicate();
	for (i = 0; i < COLM_LEN; ++i) {
		cube[1][0][i] = dup[2][0][i];
		cube[4][0][i] = dup[1][0][i];
		cube[3][0][i] = dup[4][0][i];
		cube[2][0][i] = dup[3][0][i];
	}

	cube[0][0][0] = dup[0][2][0];
	cube[0][0][2] = dup[0][0][0];
	cube[0][2][2] = dup[0][0][2];
	cube[0][2][0] = dup[0][2][2];

	cube[0][0][1] = dup[0][1][0];
	cube[0][1][2] = dup[0][0][1];
	cube[0][2][1] = dup[0][1][2];
	cube[0][1][0] = dup[0][2][1];
}

/*Up counter clockwise*/
void ui()
{
	byte_t i;

	duplicate();
	for (i = 0; i < COLM_LEN; ++i) {
		cube[2][0][i] = dup[1][0][i];
		cube[1][0][i] = dup[4][0][i];
		cube[4][0][i] = dup[3][0][i];
		cube[3][0][i] = dup[2][0][i];
	}

	cube[0][2][0] = dup[0][0][0];
	cube[0][0][0] = dup[0][0][2];
	cube[0][0][2] = dup[0][2][2];
	cube[0][2][2] = dup[0][2][0];

	cube[0][1][0] = dup[0][0][1];
	cube[0][0][1] = dup[0][1][2];
	cube[0][1][2] = dup[0][2][1];
	cube[0][2][1] = dup[0][1][0];
}

/*Down clockwise*/
void d()
{
	byte_t i;

	duplicate();
	for (i = 0; i < COLM_LEN; ++i) {
		cube[2][2][i] = dup[1][2][i];
		cube[1][2][i] = dup[4][2][i];
		cube[4][2][i] = dup[3][2][i];
		cube[3][2][i] = dup[2][2][i];
	}

	cube[5][0][0] = dup[5][2][0];
	cube[5][0][2] = dup[5][0][0];
	cube[5][2][2] = dup[5][0][2];
	cube[5][2][0] = dup[5][2][2];

	cube[5][0][1] = dup[5][1][0];
	cube[5][1][2] = dup[5][0][1];
	cube[5][2][1] = dup[5][1][2];
	cube[5][1][0] = dup[5][2][1];
}

/*Down counter clockwise*/
void di()
{
	byte_t i;

	duplicate();
	for (i = 0; i < COLM_LEN; ++i) {
		cube[1][2][i] = dup[2][2][i];
		cube[4][2][i] = dup[1][2][i];
		cube[3][2][i] = dup[4][2][i];
		cube[2][2][i] = dup[3][2][i];
	}

	cube[5][2][0] = dup[5][0][0];
	cube[5][0][0] = dup[5][0][2];
	cube[5][0][2] = dup[5][2][2];
	cube[5][2][2] = dup[5][2][0];

	cube[5][1][0] = dup[5][0][1];
	cube[5][0][1] = dup[5][1][2];
	cube[5][1][2] = dup[5][2][1];
	cube[5][2][1] = dup[5][1][0];
}

/*Get next on switch*/
void next()
{
	LCD_writeInstr(LCD_INSTR_clearDisplay);
	switch (face) {
		case 2:
			face = 3;
			look_right();
			break;
		case 3:
			face = 4;
			look_right();
			break;
		case 4:
			face = 1;
			look_right();
			break;
		case 1:
			face = 5;
			look_right();
			look_down();
			break;
		case 5:
			face = 0;
			look_down();
			look_down();
			break;
		case 0:
			face = 2;
			look_down();
			break;
		default:
			break;
	}
}
